from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
    StrictFloat
)
from typing import Union, Optional


class Arg(BaseModel):
    name: StrictStr
    default: Optional[Union[StrictStr, StrictInt, StrictBool, StrictFloat]]

    def to_string(self) -> str:

        default = self.default
        if isinstance(default, str):
            default = f'"{default}"'
    
        if default:
            return f'ARG {self.name}={default}'
        
        return f'ARG {self.name}'


from pydantic import (
    BaseModel,
    StrictStr
)

from typing import Optional


class User(BaseModel):
    user_id: StrictStr
    group_id: Optional[StrictStr]

    def to_string(self):
        if self.group_id:
            return f'USER {self.user_id}:{self.group_id}'
        
        return f'USER {self.user_id}'


import io
import tarfile
import pathlib
from typing import (
    List, 
    Union, 
    Optional,
    Callable,
    Dict,
    Any,
    Literal
)
from .layers import (
    Add,
    Arg,
    Cmd,
    Copy,
    Entrypoint,
    Env,
    Expose,
    Healthcheck,
    Label,
    Run,
    Stage,
    User,
    Volume,
    Workdir
)
from .layers.mount_types import (
    BindMount,
    CacheMount,
    SecretMount,
    SSHMount,
    TMPFSMount
)
from .memory_file import MemoryFile


class Image:

    def __init__(
        self,
        name: str,
        tag: str="latest",
        registry: str=None
    ) -> None:
        self.name = name
        self.tag = tag
        self.files: List[str] = []
        
        self.registry = registry
        self.layers: List[
            Union[
                Arg,
                Cmd,
                Copy,
                Entrypoint,
                Env,
                Expose,
                Healthcheck,
                Run,
                Stage,
                Workdir
            ]
        ] = []

        self._mount_types: Dict[
            str,
            Callable[
                [Dict[str, Any]],
                Union[
                    BindMount,
                    CacheMount,
                    SecretMount,
                    SSHMount,
                    TMPFSMount
                ]
            ]
        ] = {
            'bind': lambda config: BindMount(**config),
            'cache': lambda config: CacheMount(**config),
            'secret': lambda config: SecretMount(**config),
            'ssh': lambda config: SSHMount(**config),
            'tmpfs': lambda config: TMPFSMount(**config)
        }

    @property
    def full_name(self):
        return f'{self.name}:{self.tag}'

    def to_string(self) -> str:
        return '\n\n'.join([
            layer.to_string() for layer in self.layers
        ])
    
    def to_context(self) -> io.BytesIO:

        image_file = MemoryFile(
            self.to_string().encode()
        )

        image_file.seek(0)

        tar_file_number = image_file.file_number + 1

        tar_file = MemoryFile(
            b'', 
            file_number=tar_file_number
        )

        context = tarfile.open(
            fileobj=tar_file, 
            mode='w'
        )

        image_file_info = context.gettarinfo(
            fileobj=image_file, 
            arcname='Dockefile'
        )

        context.addfile(
            image_file_info, 
            image_file
        )

        for file in self.files:
            filename = pathlib.Path(file).name

            with open(file, "rb") as context_file:
                info = context.gettarinfo(
                    fileobj=context_file, 
                    arcname=filename
                )
                
                context.addfile(info, context_file)

        context.close()
        tar_file.seek(0)

        return tar_file
    
    def to_file(
        self,
        path: str
    ):
        self.files.append(path)

        with open(path, 'w') as dockerfile:
            dockerfile.write(
                self.to_string()
            )
    
    def add(
        self,
        source: str,
        destination: str,
        user_id: Optional[str]=None,
        group_id: Optional[str]=None,
        permissions: Optional[int]=None,
        checksum: Optional[str]=None,
        link: bool=False
    ):
        self.layers.append(
            Add(
                source=source,
                destination=destination,
                user_id=user_id,
                group_id=group_id,
                permissions=permissions,
                checksum=checksum,
                link=link
            )
        )

        return self

    def arg(
        self,
        name: str,
        default: Union[
            str,
            int,
            bool,
            float
        ]
    ):
        self.layers.append(
            Arg(
                name=name,
                default=default
            )
        )

        return self

    def cmd(
        self,
        command: List[
            Union[
                str,
                int,
                bool,
                float
            ]
        ]
    ):
        
        self.layers.append(
            Cmd(
                command=command
            )
        )

        return self
    
    def copy(
        self,
        source: str,
        destination: str,
        user_id: Optional[str]=None,
        group_id: Optional[str]=None,
        permissions: Optional[int]=None,
        from_source: Optional[str]=None,
        link: bool=False
    ):
        copy_layer = Copy(
            source=source,
            destination=destination,
            user_id=user_id,
            group_id=group_id,
            permissions=permissions,
            from_source=from_source,
            link=link
        )

        self.layers.append(copy_layer)

        self.files.append(
            f'./{copy_layer.source}'
        )

        return self
    
    def entrypoint(
        self,
        command: List[str]
    ):
        self.layers.append(
            Entrypoint(
                command=command
            )
        )

        return self
    
    def env(
        self,
        key: str,
        value: Union[
            str,
            int,
            bool,
            float
        ]
    ):
        self.layers.append(
            Env(
                key=key,
                value=value
            )
        )

    def expose(
        self,
        ports: List[int]
    ):
        self.layers.append(
            Expose(
                ports=ports
            )
        )

        return self
    
    def healthcheck(
        self,
        interval: int,
        timeout: int,
        start_period: int,
        retries: int,
        command: List[
            Union[
                str,
                int,
                bool,
                float
            ]
        ]
    ):
        self.layers.append(
            Healthcheck(
                interval=interval,
                timeout=timeout,
                start_period=start_period,
                retries=retries,
                command=Cmd(
                    command=command
                )
            )
        )

        return self
    
    def label(
        self,
        name: str,
        value: str
    ):
        self.layers.append(
            Label(
                name=name,
                value=value
            )
        )

        return self
        

    def run(
        self,
        command: str,
        network: Optional[
            Literal["default", "host", "none"]
        ]=None,
        security: Optional[
            Literal["insecure", "sandbox"]
        ]=None,
        mount: Optional[Dict[str, Any]]=None
    ):
        
        mount_type: Union[
            BindMount,
            CacheMount,
            SecretMount,
            SSHMount,
            TMPFSMount,
            None
        ] = None

        if mount:
            mount_type_name = mount.get('mount_type', 'bind')
            mount_type = self._mount_types.get(
                mount_type_name
            )(mount)

        self.layers.append(
            Run(
                command=command,
                mount=mount_type,
                network=network,
                security=security
            )
        )

        return self

    def stage(
        self,
        base: str,
        tag: str,
        alias: Optional[str]=None
    ):
        self.layers.append(
            Stage(
                base=base,
                tag=tag,
                alias=alias
            )
        )

        return self
    
    def user(
        self,
        user_id: str,
        group_id: Optional[str]=None
    ):
        self.layers.append(
            User(
                user_id=user_id,
                group_id=group_id
            )
        )

        return self
    
    def volume(
        self,
        paths: List[str]
    ):
        self.layers.append(
            Volume(
                paths=paths
            )
        )

        return self
    
    def workdir(
        self,
        path: str
    ):
        self.layers.append(
            Workdir(
                path=path
            )
        )

        return self
    


from pydantic import (
    BaseModel,
    FilePath,
    DirectoryPath,
    StrictStr,
    StrictBool,
    constr
)

from typing import Union, Optional


class Copy(BaseModel):
    source: Union[FilePath, DirectoryPath]
    destination: StrictStr
    user_id: Optional[StrictStr]
    group_id: Optional[StrictStr]
    permissions: Optional[constr(max_length=4, regex='^[0-9]*$')]
    from_source: Optional[StrictStr]
    link: StrictBool=False

    def to_string(self) -> str:
        copy_string = 'COPY'

        if self.user_id and self.group_id:
            copy_string = f'{copy_string} --chown={self.user_id}:{self.group_id}'
        
        elif self.user_id:
            copy_string = f'{copy_string} --chown={self.user_id}'

        if self.permissions:
            copy_string = f'{copy_string} --chmod={self.permissions}'
        
        if self.from_source:
            copy_string = f'{copy_string} --from={self.from_source}'

        if self.link:
            copy_string = f'{copy_string} --link'

        return f'{copy_string} ./{self.source} {self.destination}'



from pydantic import (
    BaseModel,
    StrictStr,
    conlist
)


class Volume(BaseModel):
    paths: conlist(StrictStr, min_items=1)

    def to_string(self):
        paths = ', '.join(self.paths)

        return f'VOLUME [{paths}]'


from pydantic import (
    BaseModel,
    StrictStr,
    StrictBool
)

from typing import Optional, Literal


class BindMount(BaseModel):
    mount_type: Literal["bind"]="bind"
    target: StrictStr
    source: Optional[StrictStr]
    from_source: Optional[StrictStr]
    readwrite: Optional[StrictBool]

    def to_string(self) -> str:
        mount_string = f'--mount=type={self.mount_type},target={self.target}'

        if self.source:
            mount_string = f'{mount_string},source={self.source}'

        if self.from_source:
            mount_string = f'{mount_string},from={self.from_source}'

        if self.readwrite is not None:
            readwrite_string = 'true' if self.readwrite else 'false'
            mount_string = f'{mount_string} {readwrite_string}'

        return mount_string


from pydantic import BaseModel, StrictStr
from typing import Optional, Literal, Union
from .mount_types import (
    BindMount,
    CacheMount,
    SecretMount,
    SSHMount,
    TMPFSMount
)


class Run(BaseModel):
    command: StrictStr
    mount: Optional[Union[BindMount, CacheMount, SecretMount, SSHMount, TMPFSMount]]
    network: Optional[Literal["default", "host", "none"]]
    security: Optional[Literal["insecure", "sandbox"]]

    def to_string(self) -> str:

        run_string = 'RUN'

        if self.mount:
            mount_string = self.mount.to_string()
            run_string = f'{run_string} {mount_string}'

        if self.network:
            run_string = f'{run_string} --network={self.network}'

        if self.security:
            run_string = f'{run_string} --security={self.security}'

        return f'{run_string} {self.command}'


from .add import Add
from .arg import Arg
from .cmd import Cmd
from .copy import Copy
from .entrypoint import Entrypoint
from .env import Env
from .expose import Expose
from .healthcheck import Healthcheck
from .label import Label
from .run import Run
from .stage import Stage
from .user import User
from .volume import Volume
from .workdir import Workdir


from pydantic import (
    BaseModel,
    StrictStr,
)
from typing import Union, Optional


class Label(BaseModel):
    name: StrictStr
    value: StrictStr

    def to_string(self) -> str:
        return f'ENV {self.name}="{self.value}"'


from pydantic import (
    BaseModel,
    StrictInt,
    conlist
)

from typing import List


class Expose(BaseModel):
    ports: conlist(StrictInt, min_items=1)

    def to_string(self):
        exposed_ports = ' '.join([
            str(port) for port in self.ports
        ])
        return f'EXPOSE {exposed_ports}'
        


from pydantic import (
    BaseModel,
    StrictStr
)
from typing import Optional


class Stage(BaseModel):
    base: StrictStr
    tag: StrictStr
    alias: Optional[StrictStr]

    def to_string(self) -> str:
        
        if self.alias:
            return f'FROM {self.base}:{self.tag} as {self.alias}'
        
        return f'FROM {self.base}:{self.tag}'


from pydantic import (
    BaseModel,
    StrictStr,
    StrictBool,
    constr
)
from typing import Optional, Literal


class SSHMount(BaseModel):
    mount_type: Literal["ssh"]="ssh"
    id: Optional[StrictStr]
    target: Optional[StrictStr]
    required: Optional[StrictBool]
    mode: Optional[constr(max_length=4, regex='^[0-9]*$')]
    user_id: Optional[StrictStr]
    group_id: Optional[StrictStr]

    def to_string(self) -> str:
        mount_string = f"--mount=type={self.mount_type}"

        if self.id:
            mount_string = f'{mount_string},id={self.id}'

        if self.target:
            mount_string = f'{mount_string},target={self.target}'

        if self.required is not None:
            required = "true" if self.required else False
            mount_string = f'{mount_string},required={required}'

        if self.mode:
            mount_string = f'{mount_string},mode={self.mode}'

        if self.user_id:
            mount_string = f'{mount_string},uid={self.user_id}'

        if self.group_id:
            mount_string = f'{mount_string},gid={self.group_id}'

        return mount_string


from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt
)
from typing import Optional, Literal


class TMPFSMount(BaseModel):
    mount_type: Literal["tmpfs"]="tmpfs"
    target: StrictStr
    size: Optional[StrictInt]

    def to_string(self) -> str:
        mount_string = f'--mount=type={self.mount_type},target={self.target}'

        if self.size:
            mount_string = f'{mount_string},size={self.size}'

        return mount_string


from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
    StrictFloat
)
from typing import Union, Optional


class Env(BaseModel):
    key: StrictStr
    value: Union[StrictStr, StrictInt, StrictBool, StrictFloat]

    def to_string(self) -> str:

        value = self.value
        if isinstance(value, str):
            value = f'"{value}"'

        return f'ENV {self.key}={value}'


from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
    StrictFloat
)

from typing import List, Union


class Entrypoint(BaseModel):
    command: List[StrictStr]

    def to_string(self):
        command = ', '.join([
            f'"{arg}"' for arg in self.command
        ])
        
        return f'ENTRYPOINT [{command}]'


from pydantic import (
    BaseModel,
    StrictStr,
    StrictBool,
    constr
)

from typing import Optional, Literal


class CacheMount(BaseModel):
    mount_type: Literal["cache"]="cache"
    id: StrictStr
    target: StrictStr
    source: Optional[StrictStr]
    from_source: Optional[StrictStr]
    readonly: Optional[StrictBool]
    sharing: Literal["shared", "private", "locked"]
    mode: Optional[constr(max_length=4, regex='^[0-9]*$')]
    user_id: Optional[StrictStr]
    group_id: Optional[StrictStr]

    def to_string(self) -> str:
        mount_string = f'--mount=type={self.mount_type},target={self.target}'

        if self.id:
            mount_string = f'{mount_string},id={self.id}'

        if self.source:
            mount_string = f'{mount_string},source={self.source}'

        if self.from_source:
            mount_string = f'{mount_string},from={self.from_source}'

        if self.readonly is not None:
            readonly = 'true' if self.readonly else 'false'
            mount_string = f'{mount_string},readonly={readonly}'

        if self.sharing:
            mount_string = f'{mount_string},sharing={self.sharing}'

        if self.mode:
            mount_string = f'{mount_string},mode={self.mode}'

        if self.user_id:
            mount_string = f'{mount_string},uid={self.user_id}'

        if self.group_id:
            mount_string = f'{mount_string},gid={self.group_id}'

        return mount_string


from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt
)
from .cmd import Cmd


class Healthcheck(BaseModel):
    interval: StrictInt
    timeout: StrictInt
    start_period: StrictInt
    retries: StrictInt
    command: Cmd

    def to_string(self) -> str:
        command = self.command.to_string()

        timings = f'--interval={self.interval}s --timeout={self.timeout}s --start-period={self.start_period}s'

        return f'HEALTHCHECK {timings} --retries={self.retries} {command}'